#include <QtXml>
#include <QString>

#include "doxyXmlParser.h"


static QString sect_lvls[] = {"\\chapter","\\section", "\\subsection", "\\subsubsection"};

typedef struct
{
    QString before;
    QString after;
} t_replace;

static t_replace f_replaces[] = {
    {"LPCIE_DEVINFO_FLAGS_BF_PRESENT",  "0x00000001"},
    {"LPCIE_DEVINFO_FLAGS_DAC_PRESENT", "0x00000002"},
    {"LPCIE_DEVINFO_FLAGS_GAL_PRESENT", "0x00000004"},
    {"LPCIE_DEVINFO_FLAGS_DEV_OPENED",  "0x00000100"},
    {"LPCIE_GETDEVS_FLAGS_ONLY_NOT_OPENED", "0x0001"},
    {"LPCIE_DEVNAME_SIZE", "32"},
    {"LPCIE_SERIAL_SIZE", "32"}
};

static t_replace f_id_replaces[] = {{"_1", "_"}};





static QString f_save_plain_text(QString text)
{    
    for (uint i=0; i < sizeof(f_replaces)/sizeof(f_replaces[0]); i++)
    {
        text.replace(f_replaces[i].before, f_replaces[i].after);
    }

    return text.replace("_","\\_");;
}



/* сохраняем текст, описанный через Xml как latex-текст.
   - вставляет пустые строки между параграфами
   - обрабатывает ссылки (узел типа ref)
   - note
   */
static void f_save_text(QTextStream& stream, QDomElement elem)
{
    int prev_para = 0;

    for (QDomNode node = elem.firstChild(); !node.isNull(); node = node.nextSibling())
    {
        //текст сохраняем как есть
        if (node.isText())
        {
            stream << f_save_plain_text(node.toText().data());
        }
        else if (node.isElement())
        {
            QDomElement elem = node.toElement();
            QString tag = elem.tagName();
            //ссылка на другое место
            if (elem.tagName()=="ref")
            {
                QString kind = elem.attribute("kindref");
                QString id = elem.attribute("refid");

                for (uint i=0; i < sizeof(f_id_replaces)/sizeof(f_id_replaces[0]); i++)
                {
                    id.replace(f_id_replaces[i].before, f_id_replaces[i].after);
                }

                int close = 0;
                if (kind=="member")
                {
                    stream << "\\texttt{";
                    close = 1;
                }

                stream << "\\hyperlink{" << id << "}{"
                       << f_save_plain_text(elem.text()) << "}";

                if (close)
                    stream << "}";
            }
            else if (elem.tagName() == "simplesect")
            {
                 if (elem.attribute("kind")=="note")
                 {
                     /** @todo добавить обрамление для Note */
                     f_save_text(stream, elem);
                 }
            }
            //параграф с текстом
            else if (elem.tagName()=="para")
            {
                //если до этого уже был параграф, то ставим разделитель
                if (prev_para && !elem.text().trimmed().isEmpty())
                    stream << endl << endl;
                prev_para = 1;

                f_save_text(stream, elem);
            }
        }
    }
}


static QString f_save_label(QString id)
{
    for (uint i=0; i < sizeof(f_id_replaces)/sizeof(f_id_replaces[0]); i++)
    {
        id.replace(f_id_replaces[i].before, f_id_replaces[i].after);
    }
    return "\\label{" + id  + "}";
}

static QString f_save_hypertarget(QString id)
{
    for (uint i=0; i < sizeof(f_id_replaces)/sizeof(f_id_replaces[0]); i++)
    {
        id.replace(f_id_replaces[i].before, f_id_replaces[i].after);
    }
    return "\\hypertarget{" + id  + "}{}";
}

static QString f_save_ref_text(QString text, QString id)
{
    return f_save_hypertarget(id) + f_save_plain_text(text) + f_save_label(id);
}

static QString f_save_section(QString text, QString id, int lvl)
{
    return f_save_ref_text(sect_lvls[lvl] + "{" + text.trimmed() + "}", id) + "\n";
}




static int f_save_func_descr(QTextStream& stream, doxyElemFunc func_descr, int lvl)
{
    stream << endl << f_save_section(func_descr.briefDescr(), func_descr.id(), lvl);

    stream << "\\begin{tabular}{|p{0.9\\linewidth}|} \\hline" << endl;

    stream << QString("\\textbf{Формат: \\texttt{") <<
    f_save_plain_text(func_descr.definition()) << "}}\\\\\\hline" << endl;

    // Описание
    stream << QString("\\textbf{Описание:} ");
    f_save_text(stream, func_descr.descrElem());
    stream << "\\\\\\hline" << endl;

    //Список параметров (если есть)
    if (func_descr.paramsCnt())
    {
        stream << QString("\\textbf{Параметры:} ");
        stream << "\\begin{description}" << endl;
        for (int i=0; i < func_descr.paramsCnt(); i++)
        {
            stream << "\\item[" << f_save_plain_text(
                          func_descr.param(i).name()) << "] --- ";
            f_save_text(stream, func_descr.param(i).descrElem());
        }
        stream << "\\end{description}" << endl;
        stream << "\\\\\\hline" << endl;
    }

    //Возвращаемое значение (если есть)
    if (!func_descr.returnElem().isNull())
    {
        stream << QString("\\textbf{Возвращаемое значение:}") << endl << endl;
        f_save_text(stream, func_descr.returnElem());
        stream << "\\\\\\hline" << endl;
    }


    stream << "\\end{tabular}" << endl;
    return 0;
}

static void f_save_enum_val(QTextStream& stream, doxyElemEnumVal val_descr)
{
    stream << f_save_hypertarget(val_descr.id()) <<
              "\\texttt{\\small " << f_save_plain_text(val_descr.name()) << "} &";
    //bool isInt, isHex;
    //val_descr.initializer.toInt(&isInt);
    //val_descr.initializer.toUInt(&isHex, 16);
    //if (val_descr.initializer.length() < 17)
    {
        stream << "\\texttt{\\small " << f_save_plain_text(
                      val_descr.initializer()) << "}";
    }
    stream << "&";
    f_save_text(stream, val_descr.descrElem());
    stream << f_save_label(val_descr.id()) << endl;
    stream << "\\\\\\hline" << endl;
}

static int f_save_enum_descr(QTextStream& stream, doxyElemEnum& enum_descr, int lvl)
{
    stream << endl << f_save_section(enum_descr.briefDescr(), enum_descr.id(), lvl);

    QString tableType = enum_descr.valsCount() > 10 ? "longtable" : "tabular";

    int max_char_name = 0;
    int max_char_val = 0;
    double total_len = enum_descr.valsCount() > 10 ? 0.9 : 0.87;

    for (int i=0; i < enum_descr.valsCount(); i++)
    {
         int len = enum_descr.val(i).name().length();
         if (len > max_char_name)
             max_char_name = len;
         len = f_save_plain_text(enum_descr.val(i).initializer()).length();
         if (len > max_char_val)
             max_char_val = len;
    }

    double name_width = max_char_name * 1.25/100.;
    if (name_width > 0.5)
        name_width = 0.5;
    double val_width = max_char_val * 1.25/100.;
    if (val_width>0.2)
        val_width = 0.2;
    if (val_width<0.125)
        val_width = 0.125;






    stream << "\\begin{" << tableType << "}{|m{" << name_width <<
              "\\linewidth}|m{" << val_width << "\\linewidth}|m{" <<
              total_len - name_width - val_width << "\\linewidth}|}\\hline" << endl;
    stream << "\\multicolumn{3}{|p{" << total_len << QString("\\linewidth}|}{\\textbf{Тип:} \\texttt{");
    stream << f_save_plain_text(enum_descr.name());
    stream << "}}\\\\\\hline" << endl;

    stream << "\\multicolumn{3}{|p{" << total_len << QString("\\linewidth}|}{\\textbf{Описание:}") << endl;
    f_save_text(stream, enum_descr.descrElem());
    stream << "}\\\\\\hline" << endl;

    stream << QString("\\textbf{Константа} & \\textbf{Значение} & \\textbf{Описание}\\\\\\hline");

    for (int i=0; i < enum_descr.valsCount(); i++)
    {
        f_save_enum_val(stream, enum_descr.val(i));
    }

    stream << "\\end{" << tableType << "}" << endl;
    return 0;
}

static int f_save_defines(QTextStream& stream, QList<doxyElemEnumVal> defs, int lvl)
{
    //stream << endl << sect_lvls[lvl] << QString("{Дополнительные константы.}") << endl;
    stream << "\\begin{longtable}{|m{0.45\\linewidth}|m{0.12\\linewidth}|m{0.33\\linewidth}|}\\hline" << endl;
    stream << QString("\\textbf{Константа} & \\textbf{Значение} & \\textbf{Описание}\\\\\\hline");

    for (int i=0; i < defs.length(); i++)
    {
        f_save_enum_val(stream, defs.at(i));
    }

    stream << "\\end{longtable}" << endl;
    return 0;
}


static int f_save_field_descr(QTextStream& stream, const doxyElemField *field_descr)
{
    if (!field_descr->id().isEmpty())
    {
        stream << f_save_hypertarget(field_descr->id());
    }

    stream << "\\texttt{" << f_save_plain_text(field_descr->name()) << "} &";

    f_save_text(stream, field_descr->type());
    stream << "&";

    f_save_text(stream, field_descr->descrElem());

    if (!field_descr->id().isEmpty())
    {
        stream << f_save_label(field_descr->id()) << endl;
    }
    stream << "\\\\\\hline" << endl;
    return 0;
}

static int f_save_struct(QTextStream& stream, doxyElemStruct* struct_descr, int lvl)
{
    stream << f_save_section(struct_descr->briefDescr(), struct_descr->id(), lvl);


    stream << "\\begin{tabular}{|m{0.19\\linewidth}|m{0.26\\linewidth}|m{0.35\\linewidth}|}\\hline" << endl;
    stream << QString("\\multicolumn{3}{|p{0.9\\linewidth}|}{\\textbf{Тип:} \\texttt{");
    stream << f_save_plain_text(struct_descr->name());
    stream << "}}\\\\\\hline" << endl;

    stream << QString("\\multicolumn{3}{|p{0.9\\linewidth}|}{\\textbf{Описание:}") << endl;
    f_save_text(stream, struct_descr->descrElem());
    stream << "}\\\\\\hline" << endl;

    if (struct_descr->fieldsCount())
    {
        //stream << QString("\\multicolumn{3}{|p{0.9\\linewidth}|}{\\textbf{Поля:}}\\\\\\hline");
        stream << QString("\\textbf{Поле} & \\textbf{Тип} & \\textbf{Описание поля}\\\\\\hline");
        for (int i=0; i < struct_descr->fieldsCount(); i++)
        {
            f_save_field_descr(stream, struct_descr->field(i));
        }
    }

    stream << "\\end{tabular}";
    return 0;
}


static int f_convert_tex(QTextStream& dstream, QTextStream& sstream)
{
    while (!sstream.atEnd())
    {
        QString line = sstream.readLine();
#if 0
        int ind = 0;
        do
        {
            ind = line.indexOf("\\hyperlink{");
            if (ind>=0)
            {
                line.replace(ind, 11, "\\hyperref[");
                ind = line.indexOf('}', ind);
                line.replace(ind, 1, ']');
            }
        } while (ind>=0);
#endif
        dstream << line << endl;

    }
    return 0;
}



typedef  enum
{
    FILE_PROC_TYPE_XML = 0,
    FILE_PROC_TYPE_TEX = 1
} t_file_proc_type;

class t_file_descr
{
public:
    QString srcFile() const {return m_srcFile;}
    QString destFile() const {return m_destFile;}
    int sectLvl() const {return m_lvl;}
    t_file_proc_type type() const {return m_type;}
    bool sortByTypes() const {return m_sort_by_types;}


    t_file_descr(QDomElement elem);

    t_file_descr(QString xml_file, QString tex_file, int level, t_file_proc_type type = FILE_PROC_TYPE_XML) :
        m_srcFile(xml_file), m_destFile(tex_file), m_lvl(level), m_type(type)
    {}
private:

    QString m_srcFile;
    QString m_destFile;
    int m_lvl;
    t_file_proc_type m_type;
    bool m_sort_by_types;
} ;

t_file_descr::t_file_descr(QDomElement elem)
{
    m_srcFile = elem.text().trimmed();
    m_destFile = elem.attribute("output");
    m_lvl = elem.attribute("sect_lvl", "0").toInt();
    QString type_str = elem.attribute("type", "xml");
    if (type_str=="xml")
        m_type = FILE_PROC_TYPE_XML;
    else if (type_str=="tex")
        m_type = FILE_PROC_TYPE_TEX;

    m_sort_by_types = elem.attribute("sort_by_types", "no").trimmed() == "yes" ? true : false;
}





static int f_process_tex_file(t_file_descr& fdescr)
{
    QFile src_file(fdescr.srcFile());
    if (!src_file.open(QIODevice::ReadOnly))
         return -4;
    QFile dst_file(fdescr.destFile());
    if (!dst_file.open(QIODevice::WriteOnly))
    {
        src_file.close();
        return -5;
    }

    QTextStream sstream(&src_file);
    QTextStream dstream(&dst_file);
    sstream.setCodec("UTF-8");
    dstream.setCodec("UTF-8");


    f_convert_tex(dstream, sstream);

    src_file.close();
    dst_file.close();
    return 0;
}





static int f_process_file(t_file_descr& fdescr)
{
    QList<doxyElemDescr*> clist;
    int err = parseDoxyFile(fdescr.srcFile(), clist);
    if (!err && !clist.size())
        err = -6;
    if (!err)
    {
        QFile tex(fdescr.destFile());
        if (!tex.open(QIODevice::WriteOnly))
            err = -6;
        else
        {
            QTextStream stream(&tex);
            stream.setCodec("UTF-8");

            for (int i=0; i<clist.size();i++)
            {
                doxyCompoundDescr* comp = dynamic_cast<doxyCompoundDescr*>(clist.at(i));
                doxyElemStruct* structElem = dynamic_cast<doxyElemStruct*>(clist.at(i));
                if (comp)
                {
                    QString name = comp->kind()=="group" ? comp->title() : comp->name();
                    stream << f_save_section(name, comp->id(), fdescr.sectLvl());

                    int member_lvl = fdescr.sortByTypes() ? fdescr.sectLvl()+2 : fdescr.sectLvl() + 1;

                    if (fdescr.sortByTypes())
                    {
                        QList<doxyCompoundDescr*> files = comp->fileList();
                        if (files.size())
                        {
                            stream << f_save_section(QObject::tr("Файлы."),
                                                 comp->id()+"_files", fdescr.sectLvl()+1);
                            stream << "\\begin{tabular}{|m{0.35\\linewidth}|m{0.55\\linewidth}|}\\hline" << endl;
                            //stream << "\\rowcolor{gray!10}";
                            stream << QObject::tr("Файл") << "&" << QObject::tr("Описание") <<
                                      "\\\\\\hline" << endl;

                            foreach (doxyCompoundDescr* file, files)
                            {
                                //stream << "\\rowcolor{lightgray}";
                                stream << f_save_ref_text(file->name(), file->id()) << "&";
                                f_save_text(stream, file->descrElem());
                                stream << "\\\\\\hline" << endl;
                            }

                            stream << "\\end{tabular}" << endl;
                        }
                    }


                    QList<doxyElemField*> variables = comp->variablesList();
                    if (!variables.isEmpty())
                    {
                        stream << f_save_section(QObject::tr("Глобавльные переменные"),
                                                 comp->id()+"_variables", fdescr.sectLvl()+1);

                        stream << "\\begin{tabular}{|m{0.22\\linewidth}|m{0.23\\linewidth}|m{0.43\\linewidth}|}\\hline" << endl;
                        stream << QString("\\textbf{Переменная} & \\textbf{Тип} & \\textbf{Описание}\\\\\\hline");
                        for (int i=0; i < variables.size(); i++)
                        {
                            f_save_field_descr(stream, variables.at(i));
                        }
                        stream << "\\end{tabular}";

                    }


                    QList<doxyElemEnumVal> defines = comp->definesList();
                    if (defines.size())
                    {
                        //if (fdescr.sortByTypes())
                        //{
                            stream << f_save_section(QObject::tr("Константы и макроопределения."),
                                                     comp->id()+"_defines", fdescr.sectLvl()+1);
                        //}
                        f_save_defines(stream, defines, member_lvl);
                    }

                    QList<doxyElemEnum> enums = comp->enumList();
                    if (enums.size())
                    {
                        if (fdescr.sortByTypes())
                        {
                            stream << f_save_section(QObject::tr("Перечисления."),
                                                     comp->id()+"_enums", fdescr.sectLvl()+1);
                        }

                        foreach (doxyElemEnum enumElem, enums)
                        {
                            f_save_enum_descr(stream, enumElem, member_lvl);
                        }
                    }

                    QList<doxyElemFunc> funcs = comp->funcList();
                    if (funcs.size())
                    {
                        if (fdescr.sortByTypes())
                        {
                            stream << f_save_section(QObject::tr("Функции."),
                                                     comp->id()+"_func", fdescr.sectLvl()+1);
                        }

                        foreach (doxyElemFunc funcElem, funcs)
                        {
                            f_save_func_descr(stream, funcElem, member_lvl);
                        }
                    }

                    QList<doxyElemStruct*> structs = comp->structList();
                    if (structs.size())
                    {
                        if (fdescr.sortByTypes())
                        {
                            stream << f_save_section(QObject::tr("Структуры."),
                                                     comp->id()+"_structs", fdescr.sectLvl()+1);
                        }

                        foreach (doxyElemStruct* structElem, structs)
                        {
                            f_save_struct(stream, structElem, member_lvl);
                        }
                    }
                }
                else if (structElem)
                {
                    f_save_struct(stream, structElem, fdescr.sectLvl());
                }
            }

            tex.close();
        }
    }

    qDeleteAll(clist);
    return err;
}



int main(int argc, char *argv[])
{
    QTextCodec::setCodecForCStrings(QTextCodec::codecForName("UTF-8"));
    QTextCodec::setCodecForTr(QTextCodec::codecForName("UTF-8"));

    QList<t_file_descr> fileList;

#if 0

    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__const__list.xml", "constants.tex", 1);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__type__list.xml", "types.tex", 1);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/structt__l502__cbr__coef.xml", "struct_l502_cbr_coef.tex", 2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/structt__l502__cbr.xml", "struct_l502_cbr.tex", 2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/structt__l502__info.xml", "struct_l502_info.tex", 2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__func__open.xml", "funcs_open.tex",2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__func__hnd.xml", "funcs_hnd.tex",2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__func__config.xml", "funcs_config.tex",2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__func__async.xml", "funcs_async.tex",2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__func__streams.xml", "funcs_streams.tex",2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__func__dsp.xml", "funcs_dsp.tex",2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__func__flash.xml", "funcs_flash.tex",2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/xml/group__func__misc.xml", "funcs_misc.tex",2);
    fileList << t_file_descr("/home/melkor/PRJ/L502/C/builds/l502api/docs/latex/sect_gen_descr.tex", "sect_gen_descr.tex",2, FILE_PROC_TYPE_TEX);
#endif

    if (argc < 2)
    {
        return -1;
    }

    QDomDocument doc;
    QFile file(argv[1]);
    if (!file.open(QIODevice::ReadOnly))
         return -2;

    if (!doc.setContent(&file))
    {
         file.close();
         return -3;
     }
     file.close();

     if (doc.documentElement().tagName()!="doxyXmlToLatexConfig")
     {
         return -4;
     }

     for (QDomElement elem = doc.documentElement().firstChildElement();
          !elem.isNull(); elem = elem.nextSiblingElement())
     {
         if (elem.tagName() == "file")
         {
             fileList << t_file_descr(elem);
         }
     }

    foreach (t_file_descr filedescr, fileList)
    {
        switch (filedescr.type())
        {
            case FILE_PROC_TYPE_XML:
                f_process_file(filedescr);
                break;
            case FILE_PROC_TYPE_TEX:
                f_process_tex_file(filedescr);
                break;
        }


    }
    return 0;



}


