#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "firm_check.h"
#include "burn_intf.h"
#include "lcrypt/sha.h"
#include "lcrypt/lcrypt.h"
#include "radix64.h"

#include <libxml/parser.h>
#include <libxml/tree.h>

#define LFRIM_VERSION_SIZE      32
#define LFIRM_DEV_NAME_SIZE     32
#define LFIRM_DEV_CHIPNAME_SIZE 32
#define LFIRM_DEV_REVISION      32


struct st_lfirm_info
{
    xmlDoc *doc; /* the resulting document tree */
    xmlNode *root;
    char* msg;
    size_t data_size;
    size_t msg_size;
};


/* Получаем текст из XML-узла */
static xmlChar* f_get_node_text(xmlNode* node)
{
    xmlChar* ret = NULL;
    for (node=node->children; node; node=node->next)
    {
        if ((node->type == XML_TEXT_NODE) && !xmlIsBlankNode(node))
        {
            ret = node->content;
        }
    }
    return ret;
}

void firm_parse_clean(t_lfirm_info_ptr ptr)
{
    if (ptr!=NULL)
    {
        if (ptr->doc)
        {
            xmlFreeDoc(ptr->doc);
        }
        /*
         * Cleanup function for the XML library.
         */
        xmlCleanupParser();
        /*
         * this is to debug memory for regression tests
         */
        xmlMemoryDump();


        free(ptr);
    }
}

int firm_parse_init(const char* filename, t_lfirm_info_ptr *ptr)
{
    int err = 0;
    t_lfirm_info_ptr pinfo = malloc(sizeof(t_lfirm_info));
    xmlNode *child;

    if (pinfo==NULL)
        err = L502BURN_ERR_MEMORY_ALLOC;
    else
    {
        pinfo->doc = NULL;
        pinfo->root = NULL;
    }


    if (!err)
    {
        LIBXML_TEST_VERSION;

        /*parse the file and get the DOM */
        pinfo->doc = xmlReadFile(filename, NULL, 0);
        if (pinfo->doc == NULL)
            err = L502BURN_ERR_FIRM_PARSE;
    }


    /* получаем root-элемент и проверяем его имя */
    if (!err)
    {
        pinfo->root = xmlDocGetRootElement(pinfo->doc);
        if (strcmp((char*)pinfo->root->name, "firmware"))
            err = L502BURN_ERR_FIRM_PARSE_ROOT_ELEM;
    }



    for (child = pinfo->root->children; child; child = child->next)
    {
        if ((child->type == XML_ELEMENT_NODE) && !strcmp((char*)child->name, "data"))
        {
            pinfo->msg = (char*)f_get_node_text(child);
            pinfo->msg_size = strlen(pinfo->msg);
            if (!pinfo->msg_size || (pinfo->msg_size %4))
                err = L502BURN_ERR_FIRM_PARSE_NO_DATA;
            else
            {
                pinfo->data_size = pinfo->msg_size*3/4;
                if (pinfo->msg[pinfo->msg_size-1]=='=')
                    pinfo->data_size--;
                if (pinfo->msg[pinfo->msg_size-2]=='=')
                    pinfo->data_size--;
            }


        }
    }

    if (!err && !pinfo->msg)
        err = L502BURN_ERR_FIRM_PARSE_NO_DATA;



    if (err && pinfo)
    {
        firm_parse_clean(pinfo);
    }

    if (!err)
        *ptr = pinfo;


    return err;
}

int firm_check_device(t_lfirm_info_ptr pinfo, const char* devname, const char* chip, const char* rev)
{
    int fnd = 0;
    xmlNode *child = pinfo->root->children;

    for (; !fnd && child; child = child->next)
    {
        if ((child->type == XML_ELEMENT_NODE) && !strcmp((char*)child->name, "devices"))
        {
            xmlNode *devnode = child->children;
            for (; devnode; devnode = devnode->next)
            {
                if ((devnode->type == XML_ELEMENT_NODE) && !strcmp((char*)devnode->name, "device"))
                {
                    int ok_name=0, ok_chip=chip==NULL, ok_rev = rev==NULL;
                    xmlAttr* attr;
                    for (attr = devnode->properties; attr; attr=attr->next)
                    {
                        char* val =  (char*)attr->children->content;
                        if (!strcmp((char*)attr->name, "name"))
                        {
                            if (!strcmp(val, devname))
                                ok_name = 1;
                        }
                        else if ((chip!=NULL) && !strcmp((char*)attr->name, "chip"))
                        {
                            if (!strcmp(chip, val))
                                ok_chip = 1;
                        }
                        else if ((rev!=NULL) && !strcmp((char*)attr->name, "revision"))
                        {
                            if (!strcmp(rev, val))
                                ok_rev = 1;
                        }
                    }

                    if  (ok_name && ok_chip && ok_rev)
                        fnd = 1;
                }
            }
        }
    }

    return fnd ? 0 : L502BURN_ERR_BYTE_VEREFY;
}

char* firm_get_version(t_lfirm_info* pinfo)
{
    xmlAttr* attr;
    char* res=NULL;
    for (attr = pinfo->root->properties; !res && attr; attr=attr->next)
    {
        if (!strcmp((char*)attr->name, "version"))
        {
            res = (char*)attr->children->content;;
        }

    }
    return res;
}

int firm_get_size(t_lfirm_info* pinfo, unsigned *size)
{
    *size = pinfo->data_size;
    return 0;
}

int firm_read(t_lfirm_info* pinfo, unsigned char* buf, size_t pos, size_t size)
{
    size_t first_pos = (pos/3)*4, first_offs = (pos*4)%3;
    size_t last_offs, cont_size, buf_pos=0;
    uint8_t tmp[3];

    if (pos >= pinfo->data_size)
        return 0;

    if ((pos+size) > pinfo->data_size)
        size = pinfo->data_size-pos;

    cont_size = size;
    if (first_offs)
        cont_size -= (3-first_offs);
    last_offs = (cont_size*4)%3;
    cont_size -= last_offs;


    if (first_offs)
    {
        Radix64_Decode(&pinfo->msg[first_pos], 4, tmp);
        memcpy(buf, &tmp[first_offs], 3-first_offs);
        buf_pos+=(3-first_offs);
        first_pos+=4;
    }

    Radix64_Decode(&pinfo->msg[first_pos], cont_size*4/3, &buf[buf_pos]);
    buf_pos += cont_size;
    first_pos+=(cont_size*4/3);

    if (last_offs)
    {
        Radix64_Decode(&pinfo->msg[first_pos], 4, tmp);
        memcpy(&buf[buf_pos], &tmp[0], last_offs);
    }
    return size;
}





/* проверка подписи прошивки */
int firm_check_sign(const char* filename)
{
    FILE *f, *f_sig;
    char* signame = NULL;
    int err = 0;


    signame = malloc(strlen(filename)+5);
    if (signame==NULL)
        err = L502BURN_ERR_MEMORY_ALLOC;

    if (!err)
    {
        f = fopen(filename, "rb");
        if (f==NULL)
        {
            fprintf(stderr, "Cannot open firmware file!\n");
            err = L502BURN_ERR_FILE_OPEN;
        }
    }

    if (!err)
    {
        strcpy(signame, filename);
        strcat(signame, ".sig");
        f_sig = fopen(signame, "rb");
        if (f_sig==NULL)
        {
            fprintf(stderr, "Cannot open signature file!\n");
            err = L502BURN_ERR_FILE_OPEN;
        }
    }

    if (!err)
    {
        SHA256Context sha256;
        static uint8_t f_buf[1024];
        size_t sig_size;
        uint8_t hash[SHA256HashSize];
        SHA256Reset(&sha256);

        do
        {

            size_t rd_size = fread(f_buf, 1, sizeof(f_buf), f);
            if (ferror(f))
            {
                err = L502BURN_ERR_FILE_READ;
                fprintf(stderr, "Firmware file read error!\n");
            }
            else if (rd_size)
            {
                SHA256Input(&sha256, f_buf, rd_size);
            }
        } while (!err && !feof(f));


        if (!err)
        {
            sig_size = fread(f_buf, 1, sizeof(f_buf), f_sig);
            if (ferror(f_sig))
            {
                err = L502BURN_ERR_FILE_READ;
                fprintf(stderr, "Signature file read error!\n");
            }
            else if (!feof(f_sig))
            {
                err = L502BURN_ERR_FILE_READ;
                fprintf(stderr, "Signature file is to long!\n");
            }
        }

        if (!err)
        {
            uint16_t hash_lval;
            int i;
            uint8_t end_hash[5];

            static long_t sign[2*SIGN_SIZE + 10], dec_sign[2*SIGN_SIZE + 10], n_buf[2*RSA_N_SIZE+10];
            static long_t buf1[2*SIGN_SIZE + 10], buf2[2*SIGN_SIZE + 10];

            err = PGP_SignAnalys(f_buf, sig_size, end_hash, (uint8_t*)&hash_lval, sign);

            if (!err)
            {
                //завершение вычисления SHA
                SHA256Input(&sha256, end_hash, PGP_END_HASH_SIZE);
                //res = lboot_write_data(end_hash, g_lboot_st.size, 5);
                //g_lboot_st.size +=5;
                SHA256Result(&sha256, hash);

                //проверка корректности хеша по 2 байтам из пакета
                if (hash_lval != (((uint16_t)hash[1] << 8) + hash[0]))
                {
                    err = L502BURN_ERR_INVALID_SIGN;
                }
            }

            if (!err)
            {

                //проверяем подпись по rsa
                memcpy(n_buf, rsa_n, RSA_N_SIZE);

                vlong_fast_spow_mod(sign, SIGN_SIZE/sizeof(long_t), rsa_e, n_buf, RSA_N_SIZE/sizeof(long_t),
                                                        dec_sign, buf1, buf2);


                //проверка результата - сравниваем с нужным форматом
                for (i=0; (i < SHA256HashSize)&&!err; i++)
                {
                    if (dec_sign[SHA256HashSize - 1 - i]!=hash[i])
                    {
                        err = L502BURN_ERR_INVALID_SIGN;
                    }
                }

                for (i=SHA256HashSize; (i< (SHA256HashSize + SHA256_ASN_SIZE))&&!err; i++)
                {
                    if (dec_sign[SHA256_ASN_SIZE+SHA256HashSize+SHA256HashSize-1-i]!=SHA256_ASN[i- SHA256HashSize])
                    {
                        err = L502BURN_ERR_INVALID_SIGN;
                    }
                }

                if ((dec_sign[SHA256_ASN_SIZE+SHA256HashSize] != 0x0) ||
                        (dec_sign[SIGN_SIZE-1] != 0x0) ||
                        (dec_sign[SIGN_SIZE-2] != 0x1))
                {
                    err = L502BURN_ERR_INVALID_SIGN;
                }

                for (i=SHA256HashSize + SHA256_ASN_SIZE + 1; (i< (SIGN_SIZE -2))&&!err; i++)
                {
                    if (dec_sign[i]!=0xFF)
                    {
                        err = L502BURN_ERR_INVALID_SIGN;
                    }
                }
            }
        }
    }


    if (f)
        fclose(f);
    if (f_sig)
        fclose(f_sig);
    free(signame);

    return err;
}
