#include "l502api.h"
#include "l502_fpga_regs.h"
#include "burn_intf.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


LPCIE_EXPORT(int32_t) L502_FpgaRegWrite(t_l502_hnd hnd, uint32_t reg, uint32_t val);
LPCIE_EXPORT(int32_t) L502_FpgaRegRead(t_l502_hnd hnd, uint32_t reg, uint32_t *val);

static int f_intf_open(t_l502_burn_opt *opts);
static int f_intf_close(void);
static int f_check_dev(t_l502_burn_opt *opts);
static int f_check_version(t_l502_burn_opt *opts, const char* version);
static int f_flash_set_prot(uint8_t prot);
static int f_flash_erase_64k(uint32_t addr);
static int f_flash_write_start(uint32_t addr, const uint8_t *data, uint32_t size);
static int f_flash_write_next(const uint8_t *data, uint32_t size);
static int f_flash_write_stop(void);
static int f_flash_read(uint32_t addr, uint8_t *data, uint32_t size);
static int f_exec_for_all(t_l502_burn_opt* opts, t_exec_cmd_cb cb);

static t_l502_hnd f_hnd = NULL;
static uint32_t f_addr = 0;

t_burn_intf g_pcie_intf =
{
    "pcie",
    f_intf_open,
    f_intf_close,
    f_check_dev,
    f_check_version,
    f_flash_set_prot,
    f_flash_erase_64k,
    f_flash_write_start,
    f_flash_write_next,
    f_flash_write_stop,
    f_flash_read,
    f_exec_for_all
};



static int f_intf_open(t_l502_burn_opt* opts)
{
    int err = 0;
    f_hnd = L502_Create();
    if (f_hnd==NULL)
        err = L502BURN_ERR_MEMORY_ALLOC;
    else
    {
        err = L502_Open(f_hnd, opts->serial);
        if (!err)
        {
            t_l502_info info;
            err = L502_GetDevInfo(f_hnd, &info);
            if (!err)
            {
                printf("\nOpen device: \n   name = %s\n   serial = %s\n   fpga_ver = %d.%d\n",
                       info.name, info.serial, (info.fpga_ver>>8)&0xFF, info.fpga_ver&0xFF);
                fflush(stdout);
            }

        }
    }
    return  err;
}

static int f_intf_close(void)
{
    if (f_hnd!=NULL)
    {
        L502_Close(f_hnd);
        L502_Free(f_hnd);
        f_hnd=NULL;
    }
    return 0;
}

static int f_check_dev(t_l502_burn_opt *opts)
{
    t_l502_info info;
    char rev[10];
    int32_t err = L502_GetDevInfo(f_hnd, &info);
    if (!err)
    {
        sprintf(rev, "%d", info.board_rev);
        err = firm_check_device(opts->firm_info, "L502", "Cyclone IV GX", rev);
    }
    return err;
}

static int f_check_version(t_l502_burn_opt *opts, const char* version)
{
    t_l502_info info;
    int ver_minor, ver_major, ver;
    int32_t err = L502_GetDevInfo(f_hnd, &info);
    if (!err)
    {
        sscanf(version, "%d.%d", &ver_major, &ver_minor);
        ver = (ver_major<<8)|ver_minor;
        if (ver<info.fpga_ver)
        {
            err = L502BURN_ERR_FIRM_OLD_VERSION;
        }
    }
    return err;
}

static int f_flash_set_prot(uint8_t prot)
{
    int err = L502_FpgaRegWrite(f_hnd, L502_REGS_EEPROM_WR_STATUS_EN, 1);
    if (!err)
        err = L502_FpgaRegWrite(f_hnd, L502_REGS_EEPROM_HARD_WR_STATUS_EN, 0xA502);
    if (!err)
        err = L502_FpgaRegWrite(f_hnd, L502_REGS_EEPROM_WR_STATUS, prot);
    return err;
}

static int f_flash_erase_64k(uint32_t addr)
{
    return L502_FlashErase(f_hnd, addr, 64*1024);
}

static int f_flash_write_start(uint32_t addr, const uint8_t* data, uint32_t size)
{
    f_addr = addr+size;
    return L502_FlashWrite(f_hnd, addr, data, size);
}

static int f_flash_write_next(const uint8_t* data, uint32_t size)
{
    int err = L502_FlashWrite(f_hnd, f_addr, data, size);
    if (!err)
        f_addr+=size;
    return err;
}

static int f_flash_write_stop(void)
{
    return 0;
}

static int f_flash_read(uint32_t addr, uint8_t *data, uint32_t size)
{
    return L502_FlashRead(f_hnd, addr, data, size);
}

static int f_exec_for_all(t_l502_burn_opt* opts, t_exec_cmd_cb cb)
{
    uint32_t dev_cnt;
    int32_t res;

    /* Получаем количество модулей в системе */
    res  = L502_GetSerialList(NULL, 0, 0, &dev_cnt);
    if ((res>=0) && dev_cnt)
    {
       /* Выделяем плоский массив под dev_cnt серийных номеров размером
          dev_cnt*L502_SERIAL_SIZE */
       char (*serial_list)[L502_SERIAL_SIZE] =
            (char(*)[L502_SERIAL_SIZE]) malloc
            (dev_cnt*L502_SERIAL_SIZE);
       if (serial_list==NULL)
       {
            res = L502_ERR_MEMORY_ALLOC;
       }
       else
       {
            res = L502_GetSerialList(serial_list, dev_cnt, 0,
                                     NULL);
            if (res>0)
            {
                int32_t i, err = 0;
                /* Получено res серийных номеров */
                for (i=0; !err && (i < res); i++)
                {
                    opts->serial = serial_list[i];
                    err = cb(opts);
                }
            }
            /* Освобождаем выделенный массив под серийные номера */
            free(serial_list);
       }
    }
    return res;
}






